from functools import lru_cache
from typing import Union

import jax
import numpy as np

from .base import Generator, Generator1D


class GaussHermite1D(Generator1D):
    """
    Generator for one-dimensional Gauss-Hermite nodes.

    Provides Gauss-Hermite quadrature nodes and weights on the real line, with optional scaling and shifting to custom
    domains.
    """

    def __init__(self, mean: float = 0.0, scaling: float = 1.0) -> None:
        """
        Initialize the one-dimensional Gauss-Hermite node generator.

        Parameters
        ----------
        mean : float, optional
            Center of the node sequence. Default is 0.0.
        scaling : float, optional
            Scaling of the node sequence. Default is 1.0.
        """
        super().__init__(is_nested=False)
        self.__mean = mean
        self.__scaling = scaling
        self.__cached_call = self.__make_cached_call()
        self.__cached_get_quadrature_weights = self.__make_cached_get_quadrature_weights()

    @property
    def mean(self) -> float:
        """
        Center of the node sequence.

        Returns
        -------
        float
        """
        return self.__mean

    @property
    def scaling(self) -> float:
        """
        Scaling factor of the node sequence.

        Returns
        -------
        float
        """
        return self.__scaling

    def __make_cached_call(self):
        @lru_cache(maxsize=None)
        def cached(n):
            return self.scale(np.polynomial.hermite.hermgauss(n + 1)[0])

        return cached

    def __call__(self, n: int) -> Union[jax.Array, np.ndarray]:
        """@public
        Generate the Gauss-Hermite node sequence of length `n+1`, scaled and shifted according to `self.scaling` and
        `self.mean`.

        Parameters
        ----------
        n : int
            Maximal node index (counting starts at `0`).

        Returns
        -------
        jax.Array or np.ndarray
            Scaled Gauss-Hermite nodes.
        """
        return self.__cached_call(n)

    @staticmethod
    def __make_cached_get_quadrature_weights():
        @lru_cache(maxsize=None)
        def cached(n):
            return np.polynomial.hermite.hermgauss(n + 1)[1] / np.sqrt(np.pi)

        return cached

    def scale(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes or points from the reference domain (corresponding to `mean=0` and `scaling=1`) to the custom domain.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Reference domain nodes or points.

        Mainly intended for internal use and testing.

        Returns
        -------
        jax.Array or np.ndarray
            Scaled nodes or points.
        """
        return self.__mean + self.__scaling * x

    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes or points from the custom domain back to the reference domain (corresponding to `mean=0` and
        `scaling=1`) .

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in the custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in the reference domain.
        """
        return (x - self.__mean) / self.__scaling

    def get_random(self, n: int = 1) -> Union[jax.Array, np.ndarray]:
        r"""
        Sample random points $x \sim \mathcal{N}(\mu, \frac{\sigma}{2})$, where $\mu$ = `self.mean` and
        $\sigma$ = `self.scaling`.

        The factor $\frac{1}{2}$ ensures that for $\mu=0$ and $\sigma=1$ we sample from the Gaussian measure with
        density $f(x) = \frac{1}{\sqrt{\pi}} exp(-x^2)$, which is the measure of integration for classical
        Gauss-Hermite quadrature.

        Parameters
        ----------
        n : int, optional
            Number of random points to sample (default is 1).

        Returns
        -------
        jax.Array or np.ndarray
            Randomly sampled points.
        """
        return self.scale(np.random.randn(n) / np.sqrt(2))

    def get_quadrature_weights(self, n: int) -> Union[jax.Array, np.ndarray]:
        """
        Return quadrature weights for Gauss-Hermite nodes.

        Parameters
        ----------
        n : int
            Maximal node index (counting starts at `0`).

        Returns
        -------
        jax.Array or np.ndarray
            Weights corresponding to the first `n+1` nodes.
        """
        return self.__cached_get_quadrature_weights(n)

    def __repr__(self) -> str:
        return f"Gauss-Hermite (mean = {self.__mean}, scaling = {self.__scaling})"


class GaussHermite(Generator):
    """
    Container for multiple 1D Gauss-Hermite node sequences generators with optional scaling and shifting in each
    dimension.
    """

    def __init__(
        self, mean: Union[jax.Array, np.ndarray] = None, scaling: Union[jax.Array, np.ndarray] = None, dim: int = None
    ):
        """
        Initialize the multidimensional Gauss-Hermite node generator.

        Parameters
        ----------
        mean : Union[jax.Array, np.ndarray], optional
            Node sequence centers for each dimension. Defaults to zeros.
        scaling : Union[jax.Array, np.ndarray], optional
            Node sequence scalings for each dimension. Defaults to ones.
        dim : int, optional
            Number of dimensions. Only required if neither `mean` nor `scaling` is provided.

        Raises
        ------
        ValueError
            If the dimension cannot be inferred from inputs.
        """
        if dim is None:
            if scaling is not None:
                dim = len(scaling)
            elif mean is not None:
                dim = len(mean)
            else:
                raise ValueError("Must specify at least one of 'dim', 'mean', or 'scaling'.")

        if mean is None and scaling is None:
            Generator.__init__(self, [GaussHermite1D()] * dim)
            self.__mean = np.zeros(dim)
            self.__scaling = np.ones(dim)
        else:
            self.__mean = np.zeros(dim) if mean is None else np.asarray(mean)
            self.__scaling = np.ones(dim) if scaling is None else np.asarray(scaling)
            Generator.__init__(self, [GaussHermite1D(m, a) for m, a in zip(self.__mean, self.__scaling)])

    def scale(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map multidimensional points from the reference domain (corresponding to `mean=0` and `scaling=1` in each
        dimension) to the custom domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in reference domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in custom domain.
        """
        return self.__mean + self.__scaling * x

    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map multidimensional points from the custom domain back to the reference domain (corresponding to `mean=0` and
        `scaling=1` in each dimension).

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in reference domain.
        """
        return (x - self.__mean) / self.__scaling

    def __repr__(self) -> str:
        return f"Gauss Hermite (d = {self.dim}, mean = {self.__mean.tolist()}, scaling = {self.__scaling.tolist()})"



import numpy as np
import setup


def test_nodes_scaling():

    # test with default constructors
    for node_gen in setup.generate_nodes_default(dmin=1, dmax=3):
        n = np.random.randint(low=1, high=5)
        x = node_gen.get_random(n)

        assert x.shape == (n, node_gen.dim)
        assert np.allclose(x, node_gen.scale(node_gen.scale_back(x)))

    # test with specific domains
    for node_gen in setup.generate_nodes(n=5, dmin=2, dmax=10):
        n = np.random.randint(low=0, high=5)
        x = node_gen.get_random(n)

        if n == 0:
            assert x.shape == (node_gen.dim,)
        else:
            assert x.shape == (n, node_gen.dim)
        assert np.allclose(x, node_gen.scale(node_gen.scale_back(x)))



import numpy as np
import pytest

from smolyax import indices, nodes
from smolyax.interpolation import SmolyakBarycentricInterpolator


def setup_nodes(d: int, node_type: str, default_domain: bool) -> nodes.Generator:
    """
    Set up a `nodes.Generator` instance based on the given type and domain configuration.

    Parameters
    ----------
    d : int
        The dimension of the domain.
    node_type : str
        The type of nodes to set up. Must be one of {'leja', 'gauss'}.
    default_domain : bool
        If True, use the default domain for the specified node type.
        If False, generate a random domain.

    Returns
    -------
    nodes.Generator
        An instance of the specified node type with the configured domain.
    """
    if node_type == "leja":
        if default_domain:
            return nodes.Leja(dim=d)
        else:
            domain = np.sort(np.random.rand(d, 2), axis=1)
            return nodes.Leja(domains=domain)
    elif node_type == "gauss":
        if default_domain:
            return nodes.GaussHermite(dim=d)
        else:
            mean = np.random.randn(d)
            scaling = np.random.rand(d)
            return nodes.GaussHermite(mean, scaling)
    else:
        raise ValueError("node_type must be one of {'leja', 'gauss'}")


@pytest.mark.parametrize(
    "d, m, node_type, default_domain",
    [
        (100, 1000, "leja", True),
        (100, 1000, "leja", False),
        (100, 1000, "gauss", True),
        (100, 1000, "gauss", False),
        (10000, 10000, "leja", True),
        (10000, 10000, "leja", False),
        (10000, 10000, "gauss", True),
        (10000, 10000, "gauss", False),
    ],
)
def test_smolyak_constructor_runtime(benchmark, d, m, node_type, default_domain):
    node_gen = setup_nodes(d, node_type, default_domain)
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    t = indices.find_approximate_threshold(k, m, node_gen.is_nested)
    benchmark(lambda: SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=10))


def target_f(x, theta, r):
    return 1 / (1 + theta * np.sum(x * (np.arange(x.shape[-1]) + 2) ** (-r), axis=-1))


@pytest.mark.parametrize(
    "d, m, node_type, default_domain",
    [
        (100, 1000, "leja", True),
        (100, 1000, "leja", False),
        (100, 1000, "gauss", True),
        (100, 1000, "gauss", False),
        (10000, 10000, "leja", True),
        (10000, 10000, "leja", False),
        (10000, 10000, "gauss", True),
        (10000, 10000, "gauss", False),
    ],
)
def test_smolyak_set_f_runtime(benchmark, d, m, node_type, default_domain):
    node_gen = setup_nodes(d, node_type, default_domain)
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    t = indices.find_approximate_threshold(k, m, node_gen.is_nested)
    f = lambda x: target_f(x, 2.0, 2.0)
    benchmark(lambda: SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=1, f=f))


@pytest.mark.parametrize(
    "d, m, node_type, default_domain",
    [
        (100, 1000, "leja", True),
        (100, 1000, "leja", False),
        (100, 1000, "gauss", True),
        (100, 1000, "gauss", False),
        (10000, 10000, "leja", True),
        (10000, 10000, "leja", False),
        (10000, 10000, "gauss", True),
        (10000, 10000, "gauss", False),
    ],
)
def test_smolyak_eval(benchmark, d, m, node_type, default_domain):
    node_gen = setup_nodes(d, node_type, default_domain)
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    t = indices.find_approximate_threshold(k, m, node_gen.is_nested)
    f = lambda x: target_f(x, 2.0, 2.0)
    n_inputs = 250
    smol = SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=1, n_inputs=n_inputs, f=f)
    x = np.random.randn(n_inputs, d)
    benchmark(lambda: smol(x))



from .base import Generator, Generator1D
from .gausshermite import GaussHermite, GaussHermite1D
from .leja import Leja, Leja1D

__all__ = ["Generator1D", "Generator", "Leja1D", "Leja", "GaussHermite1D", "GaussHermite"]



import itertools as it
import string
from collections import defaultdict
from typing import Callable, Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np

from . import barycentric, indices, nodes

jax.config.update("jax_enable_x64", True)


class SmolyakBarycentricInterpolator:
    r"""
    A class implementing the Smolyak operator to interpolate high-dimensional and vector-valued functions
    $f : \mathbb R^{d_{\rm in}} \to \mathbb R^{d_{\rm out}}$ via
    $$
    I^{\Lambda_{\boldsymbol{k},t}} [f] =
    \sum \limits_{\boldsymbol{\nu} \in \Lambda_{\boldsymbol{k},t}}
    \zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}} I^\boldsymbol{\nu}[f]
    $$
    where
    $\Lambda_{\boldsymbol{k},t} \subset \mathbb N^{d_{\rm in}}$ is a potentially anisotropic
    [multi-index set](indices#indexset)
    $\zeta_{\Lambda, \boldsymbol{\nu}}$ are the Smolyak coefficients and
    $I^\boldsymbol{\nu}$ is the tensor product interpolation operator with orders $\boldsymbol{\nu}$.
    """

    @property
    def d_in(self) -> int:
        """Input dimension of target function and interpolant"""
        return self.__d_in

    @property
    def d_out(self) -> int:
        """Output dimension of target function and interpolant"""
        return self.__d_out

    @property
    def n_f_evals(self) -> int:
        """Number of function evaluations (== number of interpolation nodes) used by the interpolator"""
        return self.__n_f_evals

    @property
    def n_f_evals_new(self) -> int:
        """Number of function evaluations that were not reused from previous computations"""
        return self.__n_f_evals_new

    def __init__(
        self,
        *,
        d_out: int,
        node_gen: nodes.Generator,
        k: Sequence[float],
        t: float,
        f: Callable[[Union[jax.Array, np.ndarray]], Union[jax.Array, np.ndarray]] = None,
        n_inputs: int = None,
        memory_limit: float = 4.0,
    ) -> None:
        r"""
        Initialize the Smolyak Barycentric Interpolator.

        Parameters
        ----------
        node_gen : nodes.Generator
            Generator object that returns interpolation nodes for each dimension.
        k : Union[jax.Array, np.ndarray]
            Anisotropy weight vector $\boldsymbol{k}$ of the multi-index set. Shape `(d_in,)`.
        t : float
            Threshold $t$ that controls the size of the multi-index set.
        d_out : int
            Output dimension of the target function.
        f : Callable[[Union[jax.Array, np.ndarray]], Union[jax.Array, np.ndarray]], optional
            Target function to interpolate. While `f` can be passed at construction time, for a better control over, and
            potential reuse of, function evaluations consider calling [`set_f()`](#SmolyakBarycentricInterpolator.set_f)
            *after* construction.
        n_inputs : int, default=None
            Expected number of input samples, used to pre-compile the `__call__` method via a warm-up call.
            If `n_input` is `None`, the warm-up is skipped and `n_input` is set upon first use of `__call__`.
        memory_limit : float, optional
            Maximum memory in gigabytes to use during batched evaluation. Controls the batch size
            to stay within this memory limit. Default is 4.0.
        """
        self.__d_in = len(k)
        self.__d_out = d_out

        self.__node_gen = node_gen
        self.__is_nested = node_gen.is_nested

        self.__k = k
        self.__t = t

        self.__offset = 0
        self.__n_2_F = {}
        self.__n_2_nodes = {}
        self.__n_2_weights = {}
        self.__n_2_sorted_dims = {}
        self.__n_2_sorted_degs = {}
        self.__n_2_zetas = {}

        self.__n_f_evals = indices.nodeset_cardinality(k, t, nested=self.__is_nested)
        self.__n_f_evals_new = 0

        self.__compiled_tensor_product_evaluation = None
        self.__compiled_tensor_product_gradient = None

        self.__memory_limit = memory_limit
        self.__n_inputs = n_inputs

        if f is not None:
            self.set_f(f=f)

    def set_f(
        self,
        *,
        f: Callable[[Union[jax.Array, np.ndarray]], Union[jax.Array, np.ndarray]],
        f_evals: dict[tuple, dict[tuple, jax.Array]] = None,
    ) -> dict[tuple, dict[tuple, jax.Array]]:
        """
        Compute (or reuse pre-computed) evaluations of the target function `f` at the interpolation nodes of the
        Smolyak operator.

        Parameters
        ----------
        f : Callable[[Union[jax.Array, np.ndarray]], Union[jax.Array, np.ndarray]]
            Target function to interpolate.
        f_evals : dict, optional
            A dictionary mapping interpolation nodes to function evaluations.
            If provided, these evaluations will be reused.

        Returns
        -------
        dict
            An updated dictionary containing all computed evaluations of the target function `f`.
        """
        if f_evals is None:
            f_evals = {}

        # Caching the interpolation node for nu = (0,0,...,0) for reuse in self.set_f
        zero = np.array([g(0)[0] for g in self.__node_gen])

        # get multi-indices and smolyak coefficients (only non-zero), binned by the number of active dimensions n
        n_2_nus, self.__n_2_zetas = indices.non_zero_indices_and_zetas(self.__k, self.__t)

        for n, nus in n_2_nus.items():

            # ----- Store constant term (offset) of the Smolyak operator -----

            if n == 0:
                assert len(self.__n_2_zetas[n]) == 1

                nu = ()
                if self.__is_nested:
                    f_evals_nu = f_evals
                else:
                    f_evals_nu = f_evals.get(nu, {})
                if nu not in f_evals_nu.keys():
                    f_evals_nu[nu] = f(zero.copy())
                    self.__n_f_evals_new += 1
                if not self.__is_nested:
                    f_evals[nu] = f_evals_nu

                self.__offset = self.__n_2_zetas[n][0] * f_evals_nu[nu]
                continue

            # ----- Assemble sorted dimensions and sorted degrees -----

            nn = len(nus)  # number of multi-indices of length n
            sorted_dims = np.empty((nn, n), dtype=int)
            sorted_degs = np.empty((nn, n), dtype=int)

            for i, nu in enumerate(n_2_nus[n]):
                sorted_nu = sorted(nu, key=lambda x: x[1], reverse=True)
                sorted_dims[i], sorted_degs[i] = zip(*sorted_nu)

            # ----- Assemble nodes and weights -----

            tau = tuple(int(ti) for ti in sorted_degs.max(axis=0))  # per-dimension maximal degree tau_i
            tau_max = max(tau)
            nodes = np.zeros((nn, n, tau_max + 1), dtype=float)
            weights = np.zeros((nn, n, tau_max + 1), dtype=float)

            # for each slot t, group i's by (dim,deg) so we only gen once
            for t in range(n):
                groups: dict[tuple[int, int], list[int]] = defaultdict(list)
                for i in range(nn):
                    dim = int(sorted_dims[i, t])
                    deg = int(sorted_degs[i, t])
                    groups[(dim, deg)].append(i)

                # now for each unique (dim,deg) compute pts & wts once
                for (dim, deg), idxs in groups.items():
                    pts = self.__node_gen[dim](deg)
                    wts = barycentric.compute_weights(pts)
                    nodes[idxs, t, : deg + 1] = pts
                    weights[idxs, t, : deg + 1] = wts
                # we can do even better if we vectorize node_gen(degrees) for isotropic rules, like GH or Leja

            # ----- Assemble the array storing the functions evaluations -----

            F = np.zeros((nn,) + tuple(ti + 1 for ti in tau) + (self.__d_out,), dtype=float)

            for i, nu in enumerate(n_2_nus[n]):
                x = zero.copy()

                if self.__is_nested:
                    f_evals_nu = f_evals
                else:
                    f_evals_nu = f_evals.get(nu, {})

                F_i = F[i]
                s_i = sorted_dims[i]
                argsort_s_i = np.argsort(s_i)

                ranges = [range(k + 1) for k in sorted_degs[i]]

                for mu_degrees in it.product(*ranges):
                    mu_tuple = tuple((s_i[i], mu_degrees[i]) for i in argsort_s_i if mu_degrees[i] > 0)
                    if mu_tuple not in f_evals_nu:
                        x[s_i] = [xi_k[deg] for xi_k, deg in zip(nodes[i], mu_degrees)]
                        f_evals_nu[mu_tuple] = f(x)
                        self.__n_f_evals_new += 1
                    F_i[mu_degrees] = f_evals_nu[mu_tuple]

                if not self.__is_nested:
                    f_evals[nu] = f_evals_nu
            # save as jnp data structures
            self.__n_2_F[n] = jnp.array(np.moveaxis(F, -1, 1))
            self.__n_2_sorted_degs[n] = jnp.array(sorted_degs)
            self.__n_2_sorted_dims[n] = jnp.array(sorted_dims)
            self.__n_2_nodes[n] = jnp.array(nodes)
            self.__n_2_weights[n] = jnp.array(weights)
            self.__n_2_zetas[n] = jnp.array(self.__n_2_zetas[n])

        self.__setup_eval_functions()

        return f_evals

    def __setup_eval_functions(self) -> None:

        self.__compiled_tensor_product_evaluation = jax.jit(
            jax.vmap(barycentric.evaluate_tensor_product_interpolant, in_axes=(None, 0, 0, 0, 0, 0, 0))
        )
        self.__compiled_tensor_product_gradient = jax.jit(
            jax.vmap(barycentric.evaluate_tensor_product_gradient, in_axes=(None, 0, 0, 0, 0, 0, 0))
        )

        if self.__n_inputs is not None:
            _ = self(jax.random.uniform(jax.random.PRNGKey(0), (self.__n_inputs, self.__d_in)))

    def __validate_input(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        assert bool(self.__compiled_tensor_product_evaluation) == bool(
            self.__n_2_F
        ), "The operator has not yet been set up for a target function via `set_f`."
        x = jnp.asarray(x)
        if x.shape == (self.__d_in,):
            x = x[None, :]
        assert x.shape[1] == self.__d_in, f"{x.shape[1]} != {self.__d_in}"
        self.__n_inputs = x.shape[0]
        return x

    def __call__(self, x: Union[jax.Array, np.ndarray]) -> jax.Array:
        """@public
        Evaluate the Smolyak operator at points `x`.

        Parameters
        ----------
        x : Union[jax.Array, np.ndarray]
            Points at which to evaluate the Smolyak interpolant of the target function `f`.
            Shape: `(n_points, d_in)` or `(d_in,)`, where `n_points` is the number of evaluation points
            and `d_in` is the dimension of the input domain.

        Returns
        -------
        jax.Array
            The interpolant of the target function `f` evaluated at points `x`. Shape: `(n_points, d_out)`
        """
        x = self.__validate_input(x)
        I_Lambda_x = jnp.broadcast_to(self.__offset, (self.__n_inputs, self.__d_out))

        for n in self.__n_2_F.keys():

            n_summands = self.__n_2_F[n].shape[0]
            memory_per_summand = self.__n_inputs * self.__d_out * np.prod(self.__n_2_F[n].shape[3:]) * 8 / (1024**3)
            summands_batch_size = max(1, int(np.floor(self.__memory_limit / memory_per_summand)))

            # batched processing of tensor product interpolants with n active dimensions
            for start_s in range(0, n_summands, summands_batch_size):

                end_s = min(start_s + summands_batch_size, n_summands)
                res = self.__compiled_tensor_product_evaluation(
                    x,
                    self.__n_2_F[n][start_s:end_s],
                    self.__n_2_nodes[n][start_s:end_s],
                    self.__n_2_weights[n][start_s:end_s],
                    self.__n_2_sorted_dims[n][start_s:end_s],
                    self.__n_2_sorted_degs[n][start_s:end_s],
                    self.__n_2_zetas[n][start_s:end_s],
                )
                I_Lambda_x += jnp.sum(res, axis=0)

        return I_Lambda_x

    def gradient(self, x: Union[jax.Array, np.ndarray]) -> jax.Array:
        """
        Compute the gradient of the Smolyak interpolant at the given points.

        Parameters
        ----------
        x : Union[jax.Array, numpy.ndarray]
            Points at which to evaluate the gradient. Shape: `(n_points, d_in)`.

        Returns
        -------
        jax.Array
            Gradient of the interpolant evaluated at `x`.
            Shape: `(n_points, d_out, d_in)`.
        """
        x = self.__validate_input(x)
        J_Lambda_x = jnp.zeros((x.shape[0], self.__d_out, self.__d_in))

        for n in self.__n_2_F.keys():

            # determine the number of batches that ensures that the computation stays within the given memory limit
            n_summands = self.__n_2_F[n].shape[0]
            memory_per_summand = J_Lambda_x.size * np.prod(self.__n_2_F[n].shape[3:]) * 8 / (1024**3)
            summands_batch_size = max(1, int(np.floor(self.__memory_limit / memory_per_summand)))

            # batched processing of tensor product gradients with n active dimensions
            for start_s in range(0, n_summands, summands_batch_size):
                end_s = min(start_s + summands_batch_size, n_summands)
                res = self.__compiled_tensor_product_gradient(
                    x,
                    self.__n_2_F[n][start_s:end_s],
                    self.__n_2_nodes[n][start_s:end_s],
                    self.__n_2_weights[n][start_s:end_s],
                    self.__n_2_sorted_dims[n][start_s:end_s],
                    self.__n_2_sorted_degs[n][start_s:end_s],
                    self.__n_2_zetas[n][start_s:end_s],
                )
                J_Lambda_x += jnp.sum(res, axis=0)

        return J_Lambda_x

    def integral(self) -> jax.Array:
        """
        Compute the integral of the Smolyak interpolant. Note that this is equivalent to a Smolyak quadrature
        approximation to the integral of the target function `f`.

        Returns
        -------
        jax.Array
            Integral of the interpolant. Shape: `(d_out,)`.
        """
        # assemble quadrature weights, closely following the logic in set_f to assemble nodes and weights
        # ----------------------------------------------------------------------------
        n_2_quad_weights = {}

        for n in self.__n_2_F.keys():

            nn = len(self.__n_2_zetas[n])  # number of multi-indices of length n
            sorted_degs = self.__n_2_sorted_degs[n]
            sorted_dims = self.__n_2_sorted_dims[n]
            tau = tuple(int(ti) for ti in sorted_degs.max(axis=0))  # per-dimension maximal degree tau_i

            weights_list = [np.zeros((nn, tau_i + 1), dtype=float) for tau_i in tau]
            for t in range(n):
                groups: dict[tuple[int, int], list[int]] = defaultdict(list)
                for i in range(nn):
                    dim = int(sorted_dims[i, t])
                    deg = int(sorted_degs[i, t])
                    groups[(dim, deg)].append(i)
                for (dim, deg), idxs in groups.items():
                    wts = self.__node_gen[dim].get_quadrature_weights(deg)
                    L = len(wts)
                    weights_list[t][idxs, :L] = wts
            n_2_quad_weights[n] = [jnp.array(w) for w in weights_list]

        # jit compile and evaluate tensor product terms
        # ----------------------------------------------------------------------------
        Q_Lambda = jnp.broadcast_to(self.__offset, self.__d_out)
        for n in self.__n_2_F.keys():
            q_id = "z"
            w_ids = string.ascii_lowercase[:n]
            o_id = string.ascii_lowercase[n]
            subscripts = f"{q_id}," + ",".join(f"{q_id}{w}" for w in w_ids) + f",{q_id}{o_id}{''.join(w_ids)}->{o_id}"
            Q_Lambda += jnp.einsum(subscripts, self.__n_2_zetas[n], *n_2_quad_weights[n], self.__n_2_F[n])
        return Q_Lambda.block_until_ready()



import numpy as np
import setup

from smolyax.interpolation import SmolyakBarycentricInterpolator


def test_interpolation():
    print("\nTesting vector-valued Smolyak interpolation ...")

    for node_gen in setup.generate_nodes(n=5, dmin=1, dmax=4):

        k = sorted(np.random.uniform(low=1, high=10, size=node_gen.dim))
        k /= k[0]
        d_out = np.random.randint(low=1, high=4)
        t = np.random.uniform(low=1, high=8)
        print(f"... with k = {k}, t = {t}, d_out = {d_out},", node_gen)

        f = setup.TestPolynomial(node_gen=node_gen, k=k, t=t, d_out=d_out)

        ip = SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=d_out, f=f)

        for n in range(5):
            x = node_gen.get_random(n=np.random.randint(low=1, high=5))
            assert np.allclose(
                f(x), ip(x)
            ), f"Assertion failed with\n x = {x}\n f(x) = {f(x)}\n ip(x) = {ip(x)} @ n = {n}"


def test_quadrature():
    print("\nTesting vector-valued Smolyak quadrature ...")

    for node_gen in setup.generate_nodes(n=5, dmin=1, dmax=4):

        k = sorted(np.random.uniform(low=1, high=10, size=node_gen.dim))
        k /= k[0]
        d_out = np.random.randint(low=1, high=4)
        t = np.random.uniform(low=1, high=8)
        print(f"... with k = {k}, t = {t}, d_out = {d_out},", node_gen)

        f = setup.TestPolynomial(node_gen=node_gen, k=k, t=t, d_out=d_out)

        ip = SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=d_out, f=f)

        Q_ip = ip.integral()
        assert Q_ip.shape == (d_out,)

        Q_f = f.integral()
        print(f"\tIntegral of f = {Q_f}\n\tQ[f] = {Q_ip}")
        assert np.allclose(Q_ip, Q_f), f"Assertion failed with\n Q[f] = {Q_ip}\n Integral of f = {Q_f}"


def test_gradient():
    print("\nTesting gradients of the vector-valued Smolyak interpolant ...")

    for node_gen in setup.generate_nodes(n=5, dmin=1, dmax=4):

        k = sorted(np.random.uniform(low=1, high=10, size=node_gen.dim))
        k /= k[0]
        d_out = np.random.randint(low=1, high=4)
        t = np.random.uniform(low=1, high=8)
        print(f"... with k = {k}, t = {t}, d_out = {d_out},", node_gen)

        f = setup.TestPolynomial(node_gen=node_gen, k=k, t=t, d_out=d_out)

        ip = SmolyakBarycentricInterpolator(node_gen=node_gen, k=k, t=t, d_out=d_out, f=f)

        for n in range(5):
            x = node_gen.get_random(n=np.random.randint(low=1, high=5))
            grad_f = f.gradient(x)
            grad_ip = ip.gradient(x)
            assert (
                grad_f.shape == grad_ip.shape
            ), f"Assertion failed with\n x = {x}\n grad f(x) = {grad_f}\n grad ip(x) = {grad_ip} @ n = {n}"
            assert np.allclose(
                grad_f, grad_ip
            ), f"Assertion failed with\n x = {x}\n grad f(x) = {grad_f}\n grad ip(x) = {grad_ip} @ n = {n}"



r"""
# Overview

---

ℹ️ **For installation guidelines and examples see the
    [project README](https://github.com/JoWestermann/smolyax#readme).**

---

### `smolyax.interpolation`: Smolyak barycentric operator

Central class **`SmolyakBarycentricInterpolator`** constructs the interpolant to a multivariate and vector-valued
function $f$
$$
    I^{\Lambda_{\boldsymbol{k},t}} [f] =
    \sum \limits_{\boldsymbol{\nu} \in \Lambda_{\boldsymbol{k},t}}
    \zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}} I^\boldsymbol{\nu}[f]
$$
on sparse grids governed by anisotropic multi-index sets $\Lambda_{\boldsymbol{k},t}$.
Additionally supports computing the integral (which corresponds to a quadrature approximation of the integral of $f$),
as well as evaluating its gradient.

### `smolyax.indices`: Multi-index sets
Provides numba-accelerated routines to
* construct anisotropic total degree multi-index sets of the form
$$
    \Lambda_{\boldsymbol{k}, t} := \\{\boldsymbol{\nu} \in \mathbb{N}_0^{d} \ : \ \sum_{j=1}^{d} k_j \nu_j < t\\}.
$$
* compute the Smolyak coefficients $\zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}} := \sum \limits_{\boldsymbol{e}
    \in \\{0,1\\}^d : \boldsymbol{\nu}+\boldsymbol{e} \in \Lambda_{\boldsymbol{k},t}} (-1)^{|\boldsymbol{e}|}$.
* find thresholds $t$ that for given $\boldsymbol{k}$ allows to construct $\Lambda_{\boldsymbol{k}, t}$ with a
    specified target cardinality.

### `smolyax.nodes`: Interpolation node generators

Unifies different 1-D rules under an abstract `Generator1D` interface:

| Class | Domain | Nested |
|-------|--------|--------|
| `Leja1D` | bounded interval | yes |
| `GaussHermite1D` | $\mathbb R$ (Gaussian) | no |

The interface allows for affine scaling of the domains and random sampling.

A `Generator` container bundles per-dimension generators.

### `smolyax.barycentric` and `smolyax.quadrature`: Tensor-product kernels
Implements the barycentric formula in one dimension and extends it to high-dimensional tensor-product form
$I^\boldsymbol{\nu}$. Provides the accompanying expressions for integrals and gradients of the tensor-product terms.
The routines are JIT-compatible and serve as internal utilities for the high-level interpolator.
"""



"""
Utility functionality for barycentric interpolation.
"""

import string
from typing import Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np


@jax.jit
def compute_weights(nodes: Union[jax.Array, np.ndarray]) -> jax.Array:
    """
    Compute the barycentric interpolation weights corresponding to given nodes.

    Parameters
    ----------
    nodes : Union[jax.Array, np.ndarray]
        A 1D array of shape (n,) containing n one-dimensional interpolation nodes.

    Returns
    -------
    jax.Array
        The corresponding barycentric interpolation weights.
    """
    nodes = jnp.asarray(nodes)
    diffs = nodes[:, None] - nodes
    diffs = jnp.where(diffs == 0, 1, diffs)
    return jnp.prod(1 / diffs, axis=0)


def evaluate_basis_unnormalized(x: jax.Array, xi: jax.Array, w: jax.Array, nu_i: int) -> jax.Array:
    r"""
    Evaluate the barycentric basis numerator terms at given evaluation points.

    Computes $w_j / (x - \xi_j)$ for indices $j \le \nu_i$; other entries are masked to zero.
    If an evaluation point coincides with a node $\xi_j$ for $j \le \nu_i$, returns a one-hot indicator pattern
    restricted to those entries.

    Parameters
    ----------
    x : jax.Array
        Evaluation points along one dimension. Shape: `(n_points, 1)`
    xi : jax.Array
        Interpolation nodes. Shape `(m_i,)` with `m_i > nu_i`. Entries beyond `nu_i` are not used.
    w : jax.Array
        Barycentric weights corresponding to the given interpolation nodes. Shape `(m_i,)` with `m_i > nu_i`.
        Entries beyond `nu_i` are not used.
    nu_i : int
        Polynomial degree in this dimension.

    Returns
    -------
    b : jax.Array
        Barycentric numerator terms for indices $j \le \nu_i$, or a one-hot indicator pattern if `x` coincides with a
        node. Shape `(n_points, m_i)`.
    """
    diffs = x - xi
    mask_cols = jnp.arange(diffs.shape[1]) <= nu_i
    mask_zero = jnp.any((diffs == 0) & mask_cols, axis=1)
    one_hot_pattern = (diffs == 0).astype(diffs.dtype)
    w_div_diffs = jnp.divide(w, diffs)
    b = jnp.where(mask_zero[:, None], one_hot_pattern, w_div_diffs)
    return jnp.where(mask_cols[None, :], b, 0)


def evaluate_tensor_product_interpolant(
    x: jax.Array,
    F: jax.Array,
    xi_list: jax.Array,
    w_list: jax.Array,
    sorted_dims: Sequence[int],
    sorted_degs: Sequence[int],
    zeta: int,
) -> jax.Array:
    """
    Evaluate the tensor product interpolant using the barycentric formulation.

    Parameters
    ----------
    x : jax.Array
        Points at which to evaluate the tensor product interpolant of the target function `f`.
        Should be a 2D array of shape `(n_points, d_in)` where `n_points` is the number of evaluation points
        and `d_in` is the dimension of the input domain.

    F : jax.Array
        Tensors storing the evaluations of the target function `f`.
        Should be a multi-dimensional array with shape `(d_out, mu_1, mu_2, ..., mu_n)`
        where each `mu_i` corresponds to the number of points in the ith dimension.

    xi_list : jax.Array
        Interpolation nodes. A sequence of `n` stacked 1D arrays of length `max(mu_1, mu_2, ..., mu_n)`

    w_list : jax.Array
        Interpolation weights. A sequence of `n` stacked 1D arrays of length `max(mu_1, mu_2, ..., mu_n)`

    sorted_dims : Sequence[int]
        Dimensions with nonzero interpolation degree.

    sorted_degs : Sequence[int]
        Interpolation degrees per dimension.

    zeta : int
        Smolyak coefficient.

    Returns
    -------
    jax.Array
        The evaluated tensor product interpolant at the points specified by `x`.
        The shape of the output will be `(n_points, d_out)`.
    """
    bs = []
    for i, si, nui, ni, wi in zip(F.shape[1:], sorted_dims, sorted_degs, xi_list, w_list):
        b = evaluate_basis_unnormalized(x[:, [si]], ni[:i], wi[:i], nui)
        bs.append(b / b.sum(axis=1)[:, None])

    F_axes = string.ascii_letters[: F.ndim]
    in_axis = string.ascii_letters[F.ndim]
    subscripts = f"{','.join([F_axes]+[in_axis+F_axis for F_axis in F_axes[1:]])}->{in_axis}{F_axes[0]}"

    return zeta * jnp.einsum(subscripts, F, *bs)


def evaluate_basis_gradient_unnormalized(x: jax.Array, xi: jax.Array, w: jax.Array, nu_i: int) -> jax.Array:
    r"""
    Evaluate the gradient of the barycentric basis numerator terms at given evaluation points.

    Computes $-w_j / (x - \xi_j)^2$ for indices $j \le \nu_i$; other entries are masked to zero.
    If an evaluation point coincides with a node $\xi_j$ for $j \le \nu_i$, returns `NaN` at that position.

    Parameters
    ----------
    x : jax.Array
        Evaluation points along one dimension. Shape: `(n_points, 1)`
    xi : jax.Array
        Interpolation nodes. Shape `(m_i,)` with `m_i > nu_i`. Entries beyond `nu_i` are not used.
    w : jax.Array
        Barycentric weights corresponding to the given interpolation nodes. Shape `(m_i,)` with `m_i > nu_i`.
        Entries beyond `nu_i` are not used.
    nu_i : int
        Polynomial degree in this dimension.

    Returns
    -------
    jax.Array
        Gradient values for indices $j \le \nu_i$, or `NaN` where `x` coincides with a node. Shape: `(n_points, m_i)`
    """
    diffs = (x - xi) ** 2
    mask_cols = jnp.arange(diffs.shape[1]) <= nu_i
    singular_mask = (diffs == 0) & mask_cols
    w_div_diffs = jnp.divide(-w, diffs)
    w_div_diffs = jnp.where(singular_mask, jnp.nan, w_div_diffs)
    return jnp.where(mask_cols[None, :], w_div_diffs, 0)


def evaluate_tensor_product_gradient(
    x: jax.Array,
    F: jax.Array,
    xi_list: jax.Array,
    w_list: jax.Array,
    sorted_dims: Sequence[int],
    sorted_degs: Sequence[int],
    zeta: int,
) -> jax.Array:
    """
    Evaluate the gradient of the tensor product interpolant using the barycentric formulation.

    Parameters
    ----------
    x : jax.Array
        Points at which to evaluate the tensor product interpolant of the target function `f`.
        Should be a 2D array of shape `(n_points, d_in)` where `n_points` is the number of evaluation points
        and `d_in` is the dimension of the input domain.

    F : jax.Array
        Tensors storing the evaluations of the target function `f`.
        Should be a multi-dimensional array with shape `(d_out, mu_1, mu_2, ..., mu_n)`
        where each `mu_i` corresponds to the number of points in the ith dimension.

    xi_list : jax.Array
        Interpolation nodes. A sequence of `n` stacked 1D arrays of length `max(mu_1, mu_2, ..., mu_n)`

    w_list : jax.Array
        Interpolation weights. A sequence of `n` stacked 1D arrays of length `max(mu_1, mu_2, ..., mu_n)`

    sorted_dims : Sequence[int]
        Dimensions with nonzero interpolation degree.

    sorted_degs : Sequence[int]
        Interpolation degrees per dimension.

    zeta : int
        Smolyak coefficient.

    Returns
    -------
    jax.Array
        The gradient of the tensor product interpolant evaluated at the points specified by `x`.
        The shape of the output will be `(n_points, d_out, d_in)`.
    """
    n = len(sorted_dims)
    d_out = F.shape[0]
    d_in = x.shape[1]
    n_points = x.shape[0]

    Bs = []
    for i, (si, nui, ni, wi) in enumerate(zip(sorted_dims, sorted_degs, xi_list, w_list)):

        idx = F.shape[i + 1]
        b = evaluate_basis_unnormalized(x[:, [si]], ni[:idx], wi[:idx], nui)
        b_norm = jnp.sum(b, axis=1)[:, None]

        db = evaluate_basis_gradient_unnormalized(x[:, [si]], ni[:idx], wi[:idx], nui)
        db_norm = jnp.sum(db, axis=1)[:, None]

        b = b / b_norm
        B_i = db / b_norm - b * db_norm / b_norm
        B = jnp.tile(b[:, None, :], (1, n, 1))
        B = B.at[:, i, :].set(B_i)
        Bs.append(B)

    dims = string.ascii_letters[3 : n + 3]
    subscripts = ",".join([f"b{''.join(dims)}"] + [f"ac{j}" for j in dims]) + "->abc"
    F = jnp.einsum(subscripts, F, *Bs)

    result = jnp.zeros((n_points, d_out, d_in)).at[:, :, sorted_dims].set(F)
    return zeta * result



from abc import ABC, abstractmethod
from typing import Iterator, List, Union

import jax
import numpy as np


class Generator1D(ABC):
    """
    Abstract base class for one-dimensional node sequences.

    These sequences can be used for interpolation or quadrature. Nodes are typically
    defined on a reference domain, but `scale` and `scale_back` allow mapping to custom domains.
    Supports random sampling from the measure of integration associated with the domain.
    """

    def __init__(self, is_nested: bool) -> None:
        """
        Initialize the node sequence generator.

        Parameters
        ----------
        is_nested : bool
            Whether the node sequence is nested.
        """
        self.__is_nested = is_nested

    @property
    def is_nested(self) -> bool:
        """
        Whether the node sequence is nested.

        Returns
        -------
        bool
        """
        return self.__is_nested

    @abstractmethod
    def __call__(self, n: int) -> Union[jax.Array, np.ndarray]:
        """@public
        Generate a node sequence of length `n+1`.

        Parameters
        ----------
        n : int
            Maximal node index (counting starts at `0`).

        Returns
        -------
        jax.Array or np.ndarray
            The generated node sequence.
        """
        ...

    @abstractmethod
    def scale(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes and points from the reference domain to a custom domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points on the reference domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points mapped to the custom domain.
        """
        ...

    @abstractmethod
    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes and points from the custom domain back to the reference domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points on the custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points mapped to the reference domain.
        """
        ...

    @abstractmethod
    def get_random(self, n: int = 1) -> Union[jax.Array, np.ndarray]:
        """
        Sample points randomly from the measure associated with the domain.

        This measure corresponds, for example, to the integration measure for quadrature.

        Parameters
        ----------
        n : int, optional
            Number of random points to sample (default is 1).

        Returns
        -------
        jax.Array or np.ndarray
            Randomly sampled points.
        """
        ...

    @abstractmethod
    def get_quadrature_weights(self, n: int) -> Union[jax.Array, np.ndarray]:
        """
        Return the quadrature weights for the node sequence.

        Parameters
        ----------
        n : int
            Weights corresponding to the first `n+1` nodes.

        Returns
        -------
        jax.Array or np.ndarray
            Quadrature weights.
        """
        ...


class Generator:
    """
    Container for multiple 1D node sequences generators, to be used for creating multi-dimensional sparse grids.

    Supports scaling to custom domains and random sampling.
    """

    def __init__(self, node_gens: List[Generator1D]):
        """
        Initialize the generator.

        Parameters
        ----------
        node_gens : list of Generator1D
            List of 1D node sequence generators.
        """
        assert all(g.is_nested == node_gens[0].is_nested for g in node_gens)
        self.__is_nested = node_gens[0].is_nested
        self.__dim = len(node_gens)
        self.__gens = node_gens

    @property
    def dim(self) -> int:
        """
        Dimensionality of the generator.

        Returns
        -------
        int
            Number of dimensions.
        """
        return self.__dim

    @property
    def is_nested(self) -> bool:
        """
        Whether all node sequences are nested.

        Returns
        -------
        bool
        """
        return self.__is_nested

    def __getitem__(self, i: int) -> Generator1D:
        """
        Get the 1D node generator associated with dimension `i`.

        Parameters
        ----------
        i : int
            Dimension index of the 1D generator.

        Returns
        -------
        Generator1D
        """
        return self.__gens[i]

    def __iter__(self) -> Iterator[Generator1D]:
        """
        Iterate over the 1D node generators.

        Returns
        -------
        iterator of Generator1D
        """
        for index in range(self.__dim):
            yield self[index]

    def get_random(self, n: int = 0) -> Union[jax.Array, np.ndarray]:
        """
        Generate random points from the multi-dimensional measure.

        Parameters
        ----------
        n : int, optional
            Number of random points. Both `n=0` and `n=1` will return a single point, the latter will add a separate
            axis to be consistent with the return shape for multiple points. Default is `0`.

        Returns
        -------
        jax.Array or np.ndarray
            Random points.
        """
        if n == 0:
            return np.squeeze([g.get_random() for g in self.__gens])
        return np.array([g.get_random(n) for g in self.__gens]).T

    def scale(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Scale multi-dimensional nodes or points from reference to custom domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in reference domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in custom domain.
        """
        assert x.shape[-1] == self.dim
        if x.ndim == 1:
            return np.array([g.scale(xi) for g, xi in zip(self.__gens, x)])
        else:
            return np.array([g.scale(xi) for g, xi in zip(self.__gens, x.T)]).T

    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Scale multi-dimensional nodes or points from custom domain back to reference domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in reference domain.
        """
        assert x.shape[-1] == self.dim
        if x.ndim == 1:
            return np.array([g.scale_back(xi) for g, xi in zip(self.__gens, x)])
        else:
            return np.array([g.scale_back(xi) for g, xi in zip(self.__gens, x.T)]).T



r"""
This module contains functionality to compute $\boldsymbol{k}$-weighted anisotropic multi-index sets
$\Lambda_{\boldsymbol{k}, t}$ and related data structures and quantities.

For a given weight vector $\boldsymbol{k} \in \mathbb{R}^d$ the anisotropic multi-index set
$\Lambda_{\boldsymbol{k}, t} \subset \mathbb{N}_0^d$ is defined as
$$
\Lambda_{\boldsymbol{k}, t} := \\{\boldsymbol{\nu} \in \mathbb{N}_0^{d} \ : \ \sum_{j=1}^{d} k_j \nu_j < t\\}.
$$
where $t > 0$ is a scalar threshold parameter that allows to control the size of the multi-index set.
"""

from collections import defaultdict
from typing import Sequence

import numpy as np
from numba import njit


def indexset(k: Sequence[float], t: float) -> list[tuple[tuple[int, int]]]:
    r"""
    Generate the $\boldsymbol{k}$-weighted anisotropic multi-index set
    $\Lambda_{\boldsymbol{k}, t} \subset \mathbb{N}_0^d$ with threshold $t$.

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set. The dimension $d$ is inferred as `len(k)`.
    t : float
        Threshold parameter to control the cardinality of the multi-index set.

    Returns
    -------
    list of tuples
        A list of multi-indices that satisfy the k-weighted condition with threshold `t`. The multi-indices are given in
        a tuple-based sparse format given as $((j, \nu_j))_{j \in \\{0, \dots, d\\} : \nu_j > 0\}$.

    Notes
    -----
    * To compute the cardinality of the set efficiently without constructing it use
        [`indexset_cardinality()`](#indexset_cardinality).
    * To find a suitable threshold parameter `t` that allows to construct a `k`-weighted multi-index with a specified
        cardinality use [`find_approximate_threshold()`](#find_approximate_threshold) with `nested = True`.
    """
    d = len(k)
    stack = [(0, t, ())]  # dimension, threshold, multi-index head (entries in the first dimensions)
    result = []

    while stack:
        i, remaining_t, nu_head = stack.pop()

        # Check if the stack entry is final
        if i >= d or k[i] >= remaining_t:
            result.append(nu_head)
            continue

        # Add nu_head with nu_i = 0 on to the stack
        stack.append((i + 1, remaining_t, nu_head))

        # Add all admissible nu_head with nu_i = j on to the stack
        j = 1
        k_i = k[i]
        while j * k_i < remaining_t:
            nu_extended = nu_head + ((i, j),)
            new_t = remaining_t - j * k_i
            stack.append((i + 1, new_t, nu_extended))
            j += 1

    return result


@njit(cache=True)
def indexset_cardinality(k: Sequence[float], t: float) -> int:
    r"""
    Compute the cardinality of the $\boldsymbol{k}$-weighted anisotropic multi-index set
    $\Lambda_{\boldsymbol{k}, t} \subset \mathbb{N}_0^d$ with threshold $t$.

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set. The dimension $d$ is inferred as `len(k)`.
    t : float
        Threshold parameter to control the cardinality of the multi-index set.

    Returns
    -------
    int
        Cardinality of the multi-index set $\Lambda_{\boldsymbol{k}, t}.$

    Notes
    -----
    * The result of this method is equivalent to `len(indexset(k, t))`, but it is computed significantly more
        efficiently since the index set is not explicitly constructed and numba compilation is used."
    """
    stack = [(0, 0.0)]  # dimension, threshold
    count = 0

    while stack:
        dim_i, used_t = stack.pop()

        if dim_i >= len(k):
            count += 1
            continue

        remaining_t = t - used_t

        if dim_i + 1 < len(k) and k[dim_i + 1] < remaining_t:
            stack.append((dim_i + 1, used_t))
        else:
            count += 1

        j = 1
        while used_t + j * k[dim_i] < t:
            new_used_t = used_t + j * k[dim_i]
            stack.append((dim_i + 1, new_used_t))
            j += 1

    return count


@njit(cache=True)
def smolyak_coefficient(k: Sequence[float], d: int, rem_t: float, parity: int) -> int:
    r"""
    Computes the smolyak coefficient $\zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}} := \sum
    \limits_{\boldsymbol{e} \in \\{0,1\\}^d : \boldsymbol{\nu}+\boldsymbol{e} \in \Lambda_{\boldsymbol{k},t}}
    (-1)^{|\boldsymbol{e}|}$.

    However, instead of taking $\boldsymbol{\nu}$ and checking for every $\boldsymbol{e} \in \\{0,1\\}^d$ whether
    $\sum_{j=1}^d (\nu_j + e_j) k_j < t$, the function takes the *remaining* threshold value
    $\tilde{t}_{t, \boldsymbol{k}, \boldsymbol{\nu}} := t - \sum_{j=1}^d \nu_j k_j$.
    and then checks for all $\boldsymbol{e}$ whether $\sum_{j=1}^d e_j k_j < t_{\boldsymbol{k}, \boldsymbol{\nu}}.$

    For efficiency, the indices $\boldsymbol{e}$ are not constructed explicitly. Instead, the parity of the exponent
    $|\boldsymbol{e}|$ is tracked by bit-flips while iterating over the dimensions $j \in \\{0, \dots, d\\}.q$

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set.
    d : int
        Number of dimensions (equal to `len(k)`)
    rem_t : float
        Remaining threshold value $\tilde{t}_{t, \boldsymbol{k}, \boldsymbol{\nu}} := t - \sum_{j=1}^d \nu_j k_j$
    parity : int
        Initial parity (0 or 1).

    Returns
    -------
    int
        The Smolyak coefficient $\zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}}.$
    """
    total = 0
    stack = [(0, rem_t, parity)]
    while stack:
        i, rt, p = stack.pop()
        if i >= d:
            total += 1 - (p << 1)
            continue
        # skip‐case
        if i + 1 < d and k[i + 1] < rt:
            stack.append((i + 1, rt, p))
        else:
            total += 1 - (p << 1)
        # include‐case (one copy)
        cost = k[i]
        if cost < rt:
            stack.append((i + 1, rt - cost, p ^ 1))
    return total


def __nodeset_cardinality_nested(k: Sequence[float], t: float) -> int:
    return indexset_cardinality(k, t)


@njit(cache=True)
def __nodeset_cardinality_non_nested(k: Sequence[float], t: float) -> int:
    """
    For each nu in indexset(k,t):
    if sum((-1)**e for e in abs_e_list(k,t,nu)) != 0
    add prod(v+1 for (_,v) in nu)
    —all in one Nopython pass.
    """
    d = len(k)
    stack = [(0, t, 0, 1)]  # dimension, rem_budget, parity, prod_n
    total = 0

    while stack:
        dim_i, rem_t, parity, prod_n = stack.pop()

        # terminal skip‐branch?
        if dim_i >= d or not (dim_i + 1 < d and k[dim_i + 1] < rem_t):
            zeta = smolyak_coefficient(k, d, rem_t, parity)
            if zeta != 0:
                total += prod_n

        # now expand exactly like your original indexset
        if dim_i < d:
            # skip‐branch
            if dim_i + 1 < d and k[dim_i + 1] < rem_t:
                stack.append((dim_i + 1, rem_t, parity, prod_n))
            # include‐branches for all j≥1
            cost = k[dim_i]
            j = 1
            while cost * j < rem_t:
                new_parity = parity ^ (j & 1)
                new_prod_n = prod_n * (j + 1)
                new_rem_t = rem_t - cost * j
                stack.append((dim_i + 1, new_rem_t, new_parity, new_prod_n))
                j += 1

    return total


def non_zero_indices_and_zetas(
    k: Sequence[float], t: float
) -> tuple[defaultdict[int, list[tuple[tuple[int, int], ...]]], defaultdict[int, list[float]]]:
    r"""
    Computes the subset of multi-indices $\boldsymbol{\nu}$ in $\Lambda_{\boldsymbol{k}, t}$ that have non-zero Smolyak
    coefficient $\zeta_{\Lambda_{\boldsymbol{k},t}, \boldsymbol{\nu}}$, as well as these Smolyak coefficients, grouped
    by the sparsity level $n$ (number of non-zero entries in the multi-index $\boldsymbol{\nu}$).

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set. The dimension $d$ is inferred as `len(k)`.
    t : float
        Threshold parameter to control the cardinality of the multi-index set.

    Returns
    -------
    n2nus : defaultdict[int, list[tuple[tuple[int, int], ...]]]
        Dictionary mapping sparsity level $n$ to the list of multi-indices $\boldsymbol{\nu}$
        with $n$ non-zero entries and non-zero Smolyak coefficients.
    n2zetas : defaultdict[int, list[float]]
        Dictionary mapping sparsity level $n$ to the list of corresponding non-zero Smolyak coefficients
        $\zeta_{\Lambda_{\boldsymbol{k}, t}, \boldsymbol{\nu}}$.
    """
    d = len(k)
    n2nus, n2zetas = defaultdict(list), defaultdict(list)

    stack = [(0, t, ())]
    while stack:
        i, rem_t, nu = stack.pop()
        # terminal skip check
        if i >= d or not (i + 1 < d and k[i + 1] < rem_t):
            zeta = smolyak_coefficient(k, d, rem_t, 0)
            if zeta != 0:
                n = len(nu)
                n2nus[n].append(nu)
                n2zetas[n].append(zeta)
        # expand exactly like indexset
        if i < d:
            if i + 1 < d and k[i + 1] < rem_t:
                stack.append((i + 1, rem_t, nu))
            j = 1
            while j * k[i] < rem_t:
                stack.append((i + 1, rem_t - j * k[i], nu + ((i, j),)))
                j += 1
    return n2nus, n2zetas


def nodeset_cardinality(k: Sequence[float], t: float, nested: bool = False) -> int:
    r"""
    Compute the cardinality of the set of interpolation nodes associated with the multi-index set
    $\Lambda_{\boldsymbol{k}, t}$.

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set. The dimension $d$ is inferred as `len(k)`.
    t : float
        Threshold parameter to control the cardinality of the multi-index set.
    nested : bool
        Boolean flag specifying whether the sequence of interpolation nodes used is nested or not.

    Returns
    -------
    int
        Cardinality of the set of interpolation nodes specified by $\Lambda_{\boldsymbol{k}, t}$.

    Notes
    -----
    * If `nested = True`, then the cardinality of the set of interpolation nodes is equal to the cardinality of
        $\Lambda_{\boldsymbol{k}, t}.$
    """
    if nested:
        return __nodeset_cardinality_nested(k, t)
    else:
        return __nodeset_cardinality_non_nested(k, t)


def find_approximate_threshold(
    k: Sequence[float], m: int, nested: bool, max_iter: int = 32, accuracy: float = 0.001
) -> float:
    """
    Find the approximate threshold parameter to construct a k-weighted multi-index set such that the set of
    corresponding interpolation nodes has a cardinality of approximately `m`.

    Parameters
    ----------
    k : Sequence[float]
        Weight vector of the anisotropy of the multi-index set.
    m : int
        Target cardinality of the set of interpolation nodes.
    nested : bool
        Flag to indicate whether nested or non-nested interpolation nodes are used.
    max_iter : int, optional
        Maximal number of bisection iterations. Default is 32.
    accuracy : float, optional
        Relative tolerance within which the cardinality of the set of interpolation nodes may deviate from `m`.
        Note that the accuracy may not be reached if the maximum number of iterations `max_iter` is exhausted.
        Default is 0.001.

    Returns
    -------
    float
        Threshold parameter `t` to construct a k-weighted multi-index set of size approximately `m`.

    Notes
    -----
    * The function uses a bisection method to find the threshold parameter `t` such that the cardinality
      of the set of interpolation nodes is approximately equal to `m`.
    * When `nested` is True, the cardinality of the index set is equal to the cardinality of the set of
      interpolation nodes.
    """
    assert m > 0

    if m == 1:
        return 1

    # establish search interval
    l_interval = [1.0, 2.0]
    while nodeset_cardinality(k, l_interval[0], nested) > m:
        l_interval = [l_interval[0] / 1.2, l_interval[0]]
    while nodeset_cardinality(k, l_interval[1], nested) < m:
        l_interval = [l_interval[1], l_interval[1] * 1.2]

    # bisect search interval
    def midpoint(interval):
        return interval[0] + (interval[1] - interval[0]) / 2.0

    t_cand = midpoint(l_interval)
    m_cand = nodeset_cardinality(k, t_cand, nested)
    for _ in range(max_iter):
        if m_cand > m:
            l_interval = [l_interval[0], t_cand]
        else:
            l_interval = [t_cand, l_interval[1]]
        t_cand = midpoint(l_interval)
        m_cand = nodeset_cardinality(k, t_cand, nested)

        if np.abs(m_cand - m) / m < accuracy:
            break

    return t_cand



from functools import lru_cache
from typing import Sequence, Union

import jax
import numpy as np

from .base import Generator, Generator1D


class Leja1D(Generator1D):
    """
    Generator for one-dimensional Leja nodes.

    Produces nested Leja nodes (default on [-1, 1]), with optional scaling to arbitrary bounded domains.
    """

    __nodes = np.array([0, 1, -1, 1 / np.sqrt(2), -1 / np.sqrt(2)])

    def __init__(self, domain: Union[jax.Array, np.ndarray, Sequence[float]] = None) -> None:
        """
        Initialize the one-dimensional Leja node generator.

        Parameters
        ----------
        domain : Union[jax.Array, np.ndarray, Sequence[float]], optional
            Endpoints of the domain. Defaults to `[-1, 1]` if not specified.
        """
        super().__init__(is_nested=True)
        self.__domain = domain
        self.__reference_domain = None
        if domain is not None:
            self.__domain = np.asarray(domain)
            self.__reference_domain = (-1, 1)
        self.__cached_call = self.__make_cached_call()
        self.__cached_get_quadrature_weights = self.__make_cached_get_quadrature_weights()

    @property
    def domain(self) -> Union[jax.Array, np.ndarray, Sequence[float]]:
        """
        Domain interval of the node sequence.

        Returns
        -------
        array-like of shape (2,)
        """
        return self.__domain

    @classmethod
    def __ensure_nodes(cls, n: int):
        k = cls.__nodes.shape[0]
        if n >= k:
            cls.__nodes = np.append(cls.__nodes, np.empty((n + 1 - k,)))
            for j in range(k, n + 1):
                if j % 2 == 0:
                    cls.__nodes[j] = -cls.__nodes[j - 1]
                else:
                    cls.__nodes[j] = np.sqrt((cls.__nodes[int((j + 1) / 2)] + 1) / 2)

    def __make_cached_call(self):
        @lru_cache(maxsize=None)
        def cached(n):
            self.__ensure_nodes(n)
            return self.scale(self.__nodes[: n + 1])

        return cached

    def __call__(self, n: int) -> Union[jax.Array, np.ndarray]:
        """
        Generate Leja nodes mapped to `self.domain`.

        Parameters
        ----------
        n : int
            Maximal node index (counting starts at `0`).

        Returns
        -------
        jax.Array or np.ndarray
            Leja nodes mapped to the `self.domain`.
        """
        return self.__cached_call(n)

    def __make_cached_get_quadrature_weights(self):
        @lru_cache(maxsize=None)
        def cached(n):
            self.__ensure_nodes(n)
            quadrature_points = self.__nodes[: n + 1]
            vm_matrix = np.vstack([quadrature_points**i for i in range(n + 1)])
            rhs = np.array([(1 + (-1) ** i) / (2.0 * (i + 1)) for i in range(n + 1)])
            quadrature_weights = np.linalg.solve(vm_matrix, rhs)
            return quadrature_weights

        return cached

    def scale(
        self,
        x: Union[jax.Array, np.ndarray],
        d1: Union[jax.Array, np.ndarray, Sequence[float]] = None,
        d2: Union[jax.Array, np.ndarray, Sequence[float]] = None,
    ) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes or points from interval `d1` to interval `d2` by affine transformation.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : array-like
            Points to transform.
        d1 : array-like of shape (2,), optional
            Source interval (default: reference domain `[-1, 1]`).
        d2 : array-like of shape (2,), optional
            Target interval (default: generator's domain).

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in the target domain.
        """
        if d1 is None:
            d1 = self.__reference_domain
        if d2 is None:
            d2 = self.__domain

        assert (d1 is None) == (d2 is None)
        if d1 is None:  # no scaling if no custom domains are give
            return x

        # ensure d1, d2 have shape (2, )
        d1, d2 = np.squeeze(d1), np.squeeze(d2)
        assert d1.shape == d2.shape == (2,), f"shapes {d1.shape} and {d2.shape} do not match (2, )"
        assert d1[0] < d1[1]
        assert d2[0] < d2[1]

        # ensure x has shape (n, )
        x_shape = x.shape
        x = np.squeeze(x)
        assert x.ndim <= 1

        # ensure x in d1
        valid_lower = (x >= d1[0]) | np.isclose(x, d1[0])
        valid_upper = (x <= d1[1]) | np.isclose(x, d1[1])
        assert np.all(valid_lower), f"Assertion failed: Some values are below lower bounds\n{x[~valid_lower]}"
        assert np.all(valid_upper), f"Assertion failed: Some values are above upper bounds\n{x[~valid_upper]}"

        # scale
        x = (x - d1[0]) / (d1[1] - d1[0])
        x = x * (d2[1] - d2[0]) + d2[0]

        # Return in original shape
        return x.reshape(x_shape)

    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map points from the generator's domain back to the reference domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in the custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in the reference domain `[-1, 1]`.
        """
        return self.scale(x, d1=self.__domain, d2=self.__reference_domain)

    def get_random(self, n: int = 1):
        """
        Sample random points uniformly on the domain.

        Parameters
        ----------
        n : int, optional
            Number of random points to sample (default is 1).

        Returns
        -------
        jax.Array or np.ndarray
            Randomly sampled points.
        """
        return self.scale(np.random.uniform(-1, 1, n))

    def get_quadrature_weights(self, n: int) -> Union[jax.Array, np.ndarray]:
        """
        Return quadrature weights for the Leja nodes.

        Parameters
        ----------
        n : int
            Weights corresponding to the first `n+1` nodes.

        Returns
        -------
        jax.Array or np.ndarray
            Quadrature weights.
        """
        return self.__cached_get_quadrature_weights(n)

    def __repr__(self) -> str:
        return f"Leja (domain = {self.__domain})"


class Leja(Generator):
    """
    Container for multiple 1D Leja node generators with potentially custom domains in each dimension.
    """

    def __init__(
        self,
        *,
        domains: list[Union[jax.Array, np.ndarray, Sequence[float]]] = None,
        dim: int = None,
    ):
        """
        Initialize the multidimensional Leja node generator.

        Parameters
        ----------
        domains : list of Union[jax.Array, np.ndarray, Sequence[float]], optional
            List of 1D domains (each as a 2-element sequence) for each dimension.
            If provided, `dim` must be None.
        dim : int, optional
            The number of dimensions. If provided without `domains`, uses `[-1, 1]`
            as the default domain in each dimension.

        Raises
        ------
        ValueError
            If the dimension cannot be inferred from inputs.
        """
        self.__domains = None
        self.__reference_domains = None
        if domains is not None:
            Generator.__init__(self, [Leja1D(domain) for domain in domains])
            self.__domains = np.asarray(domains)
            self.__reference_domains = np.array([[-1, 1]] * len(domains))
        elif dim is not None:
            Generator.__init__(self, [Leja1D()] * dim)
        else:
            raise ValueError("Must specify one of 'domains' or 'dim'.")

    def scale(
        self,
        x: Union[jax.Array, np.ndarray],
        d1: Union[jax.Array, np.ndarray, Sequence[float]] = None,
        d2: Union[jax.Array, np.ndarray, Sequence[float]] = None,
    ) -> Union[jax.Array, np.ndarray]:
        """
        Map nodes or points from interval `d1` to interval `d2` by affine transformation.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : array-like
            Points to transform.
        d1 : array-like of shape (2,), optional
            Source interval (default: reference domain `[-1, 1]`).
        d2 : array-like of shape (2,), optional
            Target interval (default: generator's domain).

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points in the target domain.
        """
        if d1 is None:
            d1 = self.__reference_domains
        if d2 is None:
            d2 = self.__domains

        assert (d1 is None) == (d2 is None)
        if d1 is None:  # no scaling if no custom domains are given
            return x

        # ensure d1, d2 have shape (d, 2)
        d1, d2 = np.asarray(d1), np.asarray(d2)
        assert d1.shape == d2.shape, f"shapes {d1.shape} and {d2.shape} do not match"
        d = 1 if len(d1.shape) == 1 else d1.shape[0]
        if len(d1.shape) == 1:
            d1 = d1.reshape((1, 2))
            d2 = d2.reshape((1, 2))
        for i in range(d):
            assert d1[i, 0] < d1[i, 1]
            assert d2[i, 0] < d2[i, 1]

        # ensure x has shape (n, d)
        x = np.asarray(x)
        x_shape = x.shape
        if x_shape == ():
            x = np.array([[x]])
        else:
            if len(x.shape) == 1:
                if x.shape[0] == d:
                    x = x.reshape((1, len(x)))
                else:
                    x = x.reshape((len(x), 1))

        # ensure x in d1
        valid_lower = (x >= d1[:, 0]) | np.isclose(x, d1[:, 0])
        valid_upper = (x <= d1[:, 1]) | np.isclose(x, d1[:, 1])
        assert np.all(valid_lower), f"Assertion failed: Some values are below lower bounds\n{x[~valid_lower]}"
        assert np.all(valid_upper), f"Assertion failed: Some values are above upper bounds\n{x[~valid_upper]}"

        # check
        assert len(x.shape) == len(d1.shape) == len(d2.shape) == 2
        assert x.shape[1] == d1.shape[0] == d2.shape[0]
        assert d1.shape[1] == d2.shape[1] == 2

        # scale
        x = (x - d1[:, 0]) / (d1[:, 1] - d1[:, 0])
        x = x * (d2[:, 1] - d2[:, 0]) + d2[:, 0]

        # Return in original shape
        return x.reshape(x_shape)

    def scale_back(self, x: Union[jax.Array, np.ndarray]) -> Union[jax.Array, np.ndarray]:
        """
        Map points from the generator's domain back to the reference domain.

        Mainly intended for internal use and testing.

        Parameters
        ----------
        x : jax.Array or np.ndarray
            Nodes or points in the custom domain.

        Returns
        -------
        jax.Array or np.ndarray
            Nodes or points with each componenent mapped to the reference domain `[-1, 1]`.
        """
        return self.scale(x, d1=self.__domains, d2=self.__reference_domains)

    def __repr__(self) -> str:
        if self.__domains is not None:
            return f"Leja (d = {self.dim}, domains = {self.__domains.tolist()})"
        else:
            return f"Leja (d = {self.dim})"



import itertools as it

import indices_dense_reference_impl as indices_dense
import numpy as np
import setup

from smolyax import indices as indices_sparse


def get_random_indexsets(nested: bool = False):
    d = np.random.randint(low=1, high=6)

    a = 1.1 + 2.9 * np.random.random()  # a \in [1.1,4.)
    b = 0.1 + 1.9 * np.random.random()  # b \in [0.1,3)
    k = np.log([a + b * i for i in range(d)]) / np.log(a)

    n_t = np.random.randint(low=1, high=100)
    t = indices_sparse.find_approximate_threshold(k, n_t, nested=nested)

    isparse = indices_sparse.indexset(k, t)
    idense = indices_dense.indexset(k, t)
    print(
        f"\tConstructed {d}-dimensional multi-index sets with a={a}, b={b} and target n={n_t}. "
        + f"Sets are of cardinality {len(isparse)}."
    )
    return k, t, isparse, idense


def test_validity_of_indexsets():
    print("Testing that index sets contain the correct multi-indices and none extra.")

    for i in range(10):
        k, t, isparse, idense = get_random_indexsets(nested=(i % 2) == 0)

        for idx_dense in it.product(*[range(int(np.floor(ki)) + 2) for ki in k]):
            assert (idx_dense in idense) == (np.dot(idx_dense, k) < t), (
                f"Assertion failed with\n k = {k}, t = {t},\n idx = {idx_dense},\n idx*k = {np.dot(idx_dense, k)}, "
                + f"\n (idx in i) = {idx_dense in idense},\n np.dot(idx, k) < t = {np.dot(idx_dense, k) < t}"
            )
            idx_sparse = setup.dense_index_to_sparse(idx_dense)
            assert (idx_sparse in isparse) == (np.dot(idx_dense, k) < t), (
                f"Assertion failed with\n k = {k}, t = {t},\n idx = {idx_dense},\n idx*k = {np.dot(idx_dense, k)}, "
                + f"\n (idx in i) = {idx_sparse in isparse},\n np.dot(idx, k) < t = {np.dot(idx_dense, k) < t}"
            )


def test_equality_of_sparse_and_dense_indexsets():
    print("Testing that the sparse and dense multi-index set implementations contain the same multi-indices.")

    for i in range(10):
        k, _, isparse, idense = get_random_indexsets(nested=(i % 2) == 0)

        for nu in isparse:
            nu_dense = setup.sparse_index_to_dense(nu, dim=len(k))
            assert nu_dense in idense, nu_dense

        for nu in idense:
            nu_sparse = setup.dense_index_to_sparse(nu)
            assert nu_sparse in isparse, nu_sparse


def test_smolyak_coefficients():
    print("Testing that sparse and dense computation of the Smolyak coefficients coincide.")

    for i in range(10):
        k, t, isparse, idense = get_random_indexsets(nested=(i % 2) == 0)
        d = len(k)

        for nu_sparse in isparse:
            nu_dense = setup.sparse_index_to_dense(nu_sparse, dim=len(k))
            assert nu_dense in idense
            c_sparse = indices_sparse.smolyak_coefficient(k, d, t - np.dot(nu_dense, k), 0)
            c_dense = indices_dense.smolyak_coefficient(k, t, nu=nu_dense)
            assert c_sparse == c_dense, (
                f"Assertion failed for nu_sparse={nu_sparse}, c_sparse={c_sparse}"
                + " and nu_dense={nu_dense}, c_dense={c_dense}"
            )



import numpy as np
import pytest

from smolyax import indices


@pytest.mark.parametrize(
    "d, t, m",
    [
        (10000, 5.3, 98),
        (10000, 7.78, 1000),
        (10000, 10.317, 10003),
    ],
)
def test_indexset_runtime(benchmark, d, t, m):
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    iset = benchmark(lambda: indices.indexset(k, t))
    assert len(iset) == m


@pytest.mark.parametrize(
    "d, m, nested",
    [
        (100, 1000, True),
        (100, 1000, False),
        (10000, 10000, True),
        (10000, 10000, False),
    ],
)
def test_find_threshold_runtime(benchmark, d, m, nested):
    accuracy = 0.01 if nested else 0.1
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    t = benchmark(lambda: indices.find_approximate_threshold(k, m, nested=nested, accuracy=accuracy))
    assert np.abs(indices.nodeset_cardinality(k, t, nested=nested) - m) / m < accuracy, f"d={d}, m={m}"


@pytest.mark.parametrize(
    "d, t",
    [
        (10, 6.01),
        (10, 15.416),
        (100, 7.943),
        (10000, 10.317),
    ],
)
def test_smolyak_coeffs_runtime(benchmark, d, t):
    k = np.log([2 + i for i in range(d)]) / np.log(2)
    iset = indices.indexset(k, t)
    rem_ts = [t - np.sum([nuj * k[j] for j, nuj in nu]) for nu in iset]
    benchmark(lambda: [indices.smolyak_coefficient(k, d, rt, 0) for rt in rem_ts])
